from typing import Optional, Tuple, Any, Union

from unittest import TestCase

from clvm_rs.clvm_storage import CLVMStorage, is_clvm_storage
from clvm_rs.chia_dialect import CHIA_DIALECT
from clvm_rs.eval_error import EvalError
from clvm_rs.program import Program

A_KW, C_KW, Q_KW = [getattr(CHIA_DIALECT, _) for _ in "A_KW C_KW Q_KW".split()]

Program.set_run_unsafe_max_cost(0x7FFFFFFFFFFFFFFF)


class TestProgram(TestCase):
    def test_at(self):
        p = Program.to([10, 20, 30, [15, 17], 40, 50])

        self.assertEqual(p.first(), p.at("f"))
        self.assertEqual(Program.to(10), p.at("f"))

        self.assertEqual(p.rest(), p.at("r"))
        self.assertEqual(Program.to([20, 30, [15, 17], 40, 50]), p.at("r"))

        self.assertEqual(p.rest().rest().rest().first().rest().first(), p.at("rrrfrf"))
        self.assertEqual(Program.to(17), p.at("rrrfrf"))

        self.assertRaises(ValueError, lambda: p.at("q"))
        self.assertEqual(None, p.at("ff"))
        self.assertEqual(None, p.at("ffr"))

    def test_at_many(self):
        p = Program.to([10, 20, 30, [15, 17], 40, 50])
        self.assertEqual(p.at_many("f", "rrrfrf"), [10, 17])
        self.assertEqual(p.at_many("fff", "rrrfff"), [None, None])

    def test_replace(self):
        p1 = Program.to([100, 200, 300])
        self.assertEqual(p1.replace(f=105), Program.to([105, 200, 300]))
        self.assertEqual(p1.replace(rrf=[301, 302]), Program.to([100, 200, [301, 302]]))
        self.assertEqual(
            p1.replace(f=105, rrf=[301, 302]), Program.to([105, 200, [301, 302]])
        )
        self.assertEqual(p1.replace(f=100, r=200), Program.to((100, 200)))

    def test_replace_conflicts(self):
        p1 = Program.to([100, 200, 300])
        self.assertRaises(ValueError, lambda: p1.replace(rr=105, rrf=200))

    def test_replace_conflicting_paths(self):
        p1 = Program.to([100, 200, 300])
        self.assertRaises(ValueError, lambda: p1.replace(ff=105))

    def test_replace_bad_path(self):
        p1 = Program.to([100, 200, 300])
        self.assertRaises(ValueError, lambda: p1.replace(q=105))
        self.assertRaises(ValueError, lambda: p1.replace(rq=105))

    def test_first_rest(self):
        p = Program.to([4, 5])
        self.assertEqual(p.first(), 4)
        self.assertEqual(p.rest(), [5])
        p = Program.to(4)
        self.assertEqual(p.pair, None)
        self.assertEqual(p.first(), None)
        self.assertEqual(p.rest(), None)

    def test_simple_run(self):
        p = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`
        args = Program.fromhex("ff32ff3c80")  # `(50 60)`
        r = p.run(args)
        self.assertEqual(r, 110)

    def test_run_exception(self):
        p = Program.fromhex(
            "ff08ffff0183666f6fffff018362617280"
        )  # `(x (q . foo) (q . bar))`
        err = None
        try:
            p.run(p)
        except EvalError as ee:
            err = ee
        self.assertEqual(err.args, ("clvm raise",))
        self.assertEqual(err._sexp, ["foo", "bar"])

    def test_hash(self):
        p1 = Program.fromhex("80")
        assert hash(p1) == id(p1)

    def test_long_repr(self):
        p1 = Program.fromhex(f"c062{'61'*98}")
        assert repr(p1) == f"Program(c062{'61'*33}...616161)"


def check_idempotency(p, *args):
    curried = p.curry(*args)

    f_0, args_0 = curried.uncurry()

    assert f_0 == p
    assert len(args_0) == len(args)
    for a, a0 in zip(args, args_0):
        assert a == a0

    return curried


def test_curry_uncurry():
    PLUS = Program.fromhex("10")  # `+`

    p = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`

    curried_p = check_idempotency(p)
    assert curried_p == [A_KW, [Q_KW, PLUS, 2, 5], 1]

    curried_p = check_idempotency(p, b"dogs")
    assert curried_p == [A_KW, [Q_KW, PLUS, 2, 5], [C_KW, (Q_KW, "dogs"), 1]]

    curried_p = check_idempotency(p, 200, 30)
    assert curried_p == [
        A_KW,
        [Q_KW, PLUS, 2, 5],
        [C_KW, (Q_KW, 200), [C_KW, (Q_KW, 30), 1]],
    ]

    # passing "args" here wraps the arguments in a list
    curried_p = check_idempotency(p, 50, 60, 70, 80)
    assert curried_p == [
        A_KW,
        [Q_KW, PLUS, 2, 5],
        [
            C_KW,
            (Q_KW, 50),
            [C_KW, (Q_KW, 60), [C_KW, (Q_KW, 70), [C_KW, (Q_KW, 80), 1]]],
        ],
    ]


def test_uncurry_not_curried():
    # this function has not been curried
    plus = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`
    assert plus.uncurry() == (plus, None)


def test_uncurry():
    # this is a positive test
    # `(a (q . (+ 2 5)) (c (q . 1) 1))`
    plus = Program.fromhex("ff02ffff01ff10ff02ff0580ffff04ffff0101ff018080")
    prog = Program.fromhex("ff10ff02ff0580")  # `(+ 2 5)`
    args = Program.fromhex("01")  # `1`
    assert plus.uncurry() == (prog, [args])


def test_uncurry_top_level_garbage():
    # there's garbage at the end of the top-level list
    # `(a (q . 1) (c (q . 1) (q . 1)) (q . 0x1337))`
    plus = Program.fromhex("ff02ffff0101ffff04ffff0101ffff010180ffff0182133780")
    assert plus.uncurry() == (plus, None)


def test_uncurry_not_pair():
    # the second item in the list is expected to be a pair, with a quote
    # `(a 1 (c (q . 1) (q . 1)))`
    plus = Program.fromhex("ff02ff01ffff04ffff0101ffff01018080")
    assert plus.uncurry() == (plus, None)


def test_uncurry_args_garbage():
    # there's garbage at the end of the args list
    # `(a (q . 1) (c (q . 1) (q . 1) (q . 4919)))`
    plus = Program.fromhex("ff02ffff0101ffff04ffff0101ffff0101ffff018213378080")
    assert plus.uncurry() == (plus, None)


class SimpleStorage(CLVMStorage):
    """
    A simple implementation of `CLVMStorage`.
    """

    atom: Optional[bytes]

    def __init__(self, atom, pair):
        self.atom = atom
        self._pair = pair

    @property
    def pair(self) -> Optional[Tuple["CLVMStorage", "CLVMStorage"]]:
        return self._pair


class Uncachable(SimpleStorage):
    """
    This object does not allow `_cached_sha256_treehash` or `_cached_serialization`
    to be stored.
    """

    def get_th(self):
        return None

    def set_th(self, v):
        raise AttributeError("can't set property")

    _cached_sha256_treehash = property(get_th, set_th)
    _cached_serialization = property(get_th, set_th)


def convert_atom_to_bytes(castable: Any) -> Optional[bytes]:
    return Program.to(castable).atom


def validate_program(program):
    validate_stack = [program]
    while validate_stack:
        v = validate_stack.pop()
        assert isinstance(v, Program)
        if v.pair:
            assert isinstance(v.pair, tuple)
            v1, v2 = v.pair
            assert is_clvm_storage(v1)
            assert is_clvm_storage(v2)
            s1, s2 = v.pair
            validate_stack.append(s1)
            validate_stack.append(s2)
        else:
            assert isinstance(v.atom, bytes)


def print_leaves(tree: Program) -> str:
    a = tree.atom
    if a is not None:
        if len(a) == 0:
            return "() "
        return "%d " % a[0]

    ret = ""
    pair = tree.pair
    assert pair is not None
    for i in pair:
        ret += print_leaves(i)

    return ret


def print_tree(tree: Program) -> str:
    a = tree.atom
    if a is not None:
        if len(a) == 0:
            return "() "
        return "%d " % a[0]

    pair = tree.pair
    assert pair is not None
    ret = "("
    for i in pair:
        ret += print_tree(i)
    ret += ")"
    return ret


class ProgramTest(TestCase):
    def test_cast_1(self):
        # this was a problem in `clvm_tools` and is included
        # to prevent regressions
        program = Program.to(b"foo")
        t1 = program.to([1, program])
        validate_program(t1)

    def test_wrap_program(self):
        # it's a bit of a layer violation that CLVMStorage unwraps Program, but we
        # rely on that in a fair number of places for now. We should probably
        # work towards phasing that out
        o = Program.to(Program.to(1))
        assert o.atom == bytes([1])

    def test_arbitrary_underlying_tree(self) -> None:
        # Program provides a view on top of a tree of arbitrary types, as long as
        # those types implement the CLVMStorage protocol. This is an example of
        # a tree that's generated
        class GeneratedTree(CLVMStorage):
            depth: int = 4
            val: int = 0

            def __init__(self, depth, val):
                assert depth >= 0
                self.depth = depth
                self.val = val
                self._cached_sha256_treehash = None
                self.atom = None if self.depth > 0 else bytes([self.val])

            @property
            def pair(self) -> Optional[Tuple[Any, Any]]:
                if self.depth == 0:
                    return None
                new_depth: int = self.depth - 1
                return (
                    GeneratedTree(new_depth, self.val),
                    GeneratedTree(new_depth, self.val + 2**new_depth),
                )

        tree = Program.to(GeneratedTree(5, 0))
        assert (
            print_leaves(tree)
            == "0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 "
            + "16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 "
        )

        tree = Program.to(GeneratedTree(3, 0))
        assert print_leaves(tree) == "0 1 2 3 4 5 6 7 "

        tree = Program.to(GeneratedTree(3, 10))
        assert print_leaves(tree) == "10 11 12 13 14 15 16 17 "

        # this is just for `coverage`
        assert print_leaves(Program.to(0)) == "() "

    def test_looks_like_clvm_object(self):
        # this function can't look at the values, that would cause a cascade of
        # eager evaluation/conversion
        class dummy:
            pass

        obj = dummy()
        obj.atom = None
        obj.pair = None
        print(dir(obj))
        assert is_clvm_storage(obj)

        obj = dummy()
        obj.pair = None
        assert not is_clvm_storage(obj)

        obj = dummy()
        obj.atom = None
        assert not is_clvm_storage(obj)

    def test_list_conversions(self):
        a = Program.to([1, 2, 3])
        assert print_tree(a) == "(1 (2 (3 () )))"

    def test_string_conversions(self):
        a = Program.to("foobar")
        assert a.atom == "foobar".encode()

    def test_int_conversions(self):
        def check(v: int, h: Union[str, list]):
            a = Program.to(v)
            b = bytes.fromhex(h) if isinstance(h, str) else bytes(h)
            assert a.atom == b
            # note that this compares to the atom, not the serialization of that atom
            # so 16384 codes as 0x4000, not 0x824000

        check(1337, "0539")
        check(-128, "80")
        check(0, "")
        check(1, "01")
        check(-1, "ff")

        for v in range(1, 0x80):
            check(v, [v])

        for v in range(0x80, 0xFF):
            check(v, [0, v])

        for v in range(128):
            check(-v - 1, [255 - v])

        check(127, "7f")
        check(128, "0080")
        check(256, "0100")
        check(-256, "ff00")
        check(16384, "4000")
        check(32767, "7fff")
        check(32768, "008000")
        check(-32768, "8000")
        check(-32769, "ff7fff")

    def test_int_round_trip(self):
        def check(n):
            p = Program.to(n)
            assert int(p) == n
            assert p.int_from_bytes(p.atom) == n
            assert Program.int_to_bytes(n) == p.atom

        for n in range(0, 256):
            check(n)
            check(-n)

        for n in range(0, 65536, 97):
            check(n)
            check(-n)

    def test_empty_list_conversions(self):
        a = Program.to([])
        assert a.atom == b""

    def test_eager_conversion(self):
        with self.assertRaises(ValueError):
            Program.to(("foobar", (1, {})))

    def test_convert_atom(self):
        assert convert_atom_to_bytes(0x133742) == bytes([0x13, 0x37, 0x42])
        assert convert_atom_to_bytes(0x833742) == bytes([0x00, 0x83, 0x37, 0x42])
        assert convert_atom_to_bytes(0) == b""

        assert convert_atom_to_bytes("foobar") == "foobar".encode()
        assert convert_atom_to_bytes("") == b""

        assert convert_atom_to_bytes(b"foobar") == b"foobar"
        assert convert_atom_to_bytes([]) == b""

        assert convert_atom_to_bytes([1, 2, 3]) is None

        assert convert_atom_to_bytes((1, 2)) is None

        with self.assertRaises(ValueError):
            assert convert_atom_to_bytes({})

    def test_to_nil(self):
        self.assertEqual(Program.to([]), 0)
        self.assertEqual(Program.to(0), 0)
        self.assertEqual(Program.to(b""), 0)

    def test_tree_hash_caching(self):
        o = SimpleStorage(b"foo", None)
        eh = "0080b50a51ecd0ccfaaa4d49dba866fe58724f18445d30202bafb03e21eef6cb"
        p = Program.to(o)
        self.assertEqual(p.tree_hash().hex(), eh)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(o._cached_sha256_treehash.hex(), eh)

        o2 = SimpleStorage(None, (o, o))
        eh2 = "4a40c538671ef10c8d956e5dd3625e167c8adfb666c943f67f91ea58fd7a302c"
        p2 = Program.to(o2)
        self.assertEqual(p2.tree_hash().hex(), eh2)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)
        self.assertEqual(o._cached_sha256_treehash.hex(), eh)
        self.assertEqual(o2._cached_sha256_treehash.hex(), eh2)

        p2p = Program.to((p, p))
        self.assertEqual(p2p.tree_hash().hex(), eh2)
        self.assertEqual(p2p._cached_sha256_treehash.hex(), eh2)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)

        o3 = SimpleStorage(None, (o2, o2))
        eh3 = "280df61ed70cac1ec3cf9811c15f75e6698516b0354252960a62fa31240e4970"
        p3 = Program.to(o3)
        self.assertEqual(p3.tree_hash().hex(), eh3)
        self.assertEqual(p3._cached_sha256_treehash.hex(), eh3)
        self.assertEqual(o._cached_sha256_treehash.hex(), eh)
        self.assertEqual(o2._cached_sha256_treehash.hex(), eh2)
        self.assertEqual(o3._cached_sha256_treehash.hex(), eh3)

        p3p = Program.to((p2, p2))
        self.assertEqual(p3p.tree_hash().hex(), eh3)
        self.assertEqual(p3p._cached_sha256_treehash.hex(), eh3)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)

    def test_tree_hash_no_caching(self):
        o = Uncachable(b"foo", None)
        eh = "0080b50a51ecd0ccfaaa4d49dba866fe58724f18445d30202bafb03e21eef6cb"
        p = Program.to(o)
        self.assertEqual(p.tree_hash().hex(), eh)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(o._cached_sha256_treehash, None)

        o2 = Uncachable(None, (o, o))
        eh2 = "4a40c538671ef10c8d956e5dd3625e167c8adfb666c943f67f91ea58fd7a302c"
        p2 = Program.to(o2)
        self.assertEqual(p2.tree_hash().hex(), eh2)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)
        self.assertEqual(o._cached_sha256_treehash, None)
        self.assertEqual(o2._cached_sha256_treehash, None)

        p2p = Program.to((p, p))
        self.assertEqual(p2p.tree_hash().hex(), eh2)
        self.assertEqual(p2p._cached_sha256_treehash.hex(), eh2)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)

        o3 = Uncachable(None, (o2, o2))
        eh3 = "280df61ed70cac1ec3cf9811c15f75e6698516b0354252960a62fa31240e4970"
        p3 = Program.to(o3)
        self.assertEqual(p3.tree_hash().hex(), eh3)
        self.assertEqual(p3._cached_sha256_treehash.hex(), eh3)
        self.assertEqual(o._cached_sha256_treehash, None)
        self.assertEqual(o2._cached_sha256_treehash, None)
        self.assertEqual(o3._cached_sha256_treehash, None)

        p3p = Program.to((p2, p2))
        self.assertEqual(p3p.tree_hash().hex(), eh3)
        self.assertEqual(p3p._cached_sha256_treehash.hex(), eh3)
        self.assertEqual(p._cached_sha256_treehash.hex(), eh)
        self.assertEqual(p2._cached_sha256_treehash.hex(), eh2)

def test_repr() -> None:
    temp = Program.to([8, (1, "foo")])
    assert f"{temp}" == "ff08ffff0183666f6f80"

    Program.set_run_unsafe_max_cost(11000000000)

    try:
        temp.run([])
        assert False
    except EvalError as e:
        assert f"{e}" == "(clvm raise, 83666f6f)"



#!/usr/bin/env python3

import subprocess
import glob
import time
import sys
import platform
from colorama import init, Fore, Style
from run import run_clvm
from os.path import isfile, splitext, basename

init()
ret = 0

expect = {
    "args-add": "cost exceeded",
    "args-all": "cost exceeded",
    "args-and": "cost exceeded",
    "args-any": "cost exceeded",
    "args-cat": "cost exceeded",
    "args-mul": "cost exceeded",
    "args-or": "cost exceeded",
    "args-point_add": "cost exceeded",
    "args-sha": "cost exceeded",
    "args-sub": "cost exceeded",
    "args-unknown-1": "cost exceeded",
    "args-unknown-2": "cost exceeded",
    "args-unknown-3": "cost exceeded",
    "args-unknown-4": "cost exceeded",
    "args-unknown-5": "too many pairs",
    "args-unknown-6": "too many pairs",
    "args-unknown-7": "too many pairs",
    "args-unknown-8": "too many pairs",
    "args-unknown-9": "too many pairs",
    "args-xor": "cost exceeded",
    "recursive-add": "cost exceeded",
    "recursive-ash": "cost exceeded",
    "recursive-cat": "cost exceeded",
    "recursive-cons": "too many pairs",
    "recursive-div": "cost exceeded",
    "recursive-lsh": "cost exceeded",
    "recursive-mul": "cost exceeded",
    "recursive-not": "cost exceeded",
    "recursive-pubkey": "cost exceeded",
    "recursive-sub": "too many pairs",
    "softfork-1": "cost exceeded",
    "softfork-2": "cost exceeded",
}

for fn in glob.glob("programs/large-atom-*.hex.invalid"):
    try:
        print(fn)
        run_clvm(fn)
        ret = 1
        print(Fore.RED + "FAILED: expected parse failure" + Style.RESET_ALL)
    except Exception as e:
        print(Fore.GREEN + f"OK: expected: {e}" + Style.RESET_ALL)


for fn in glob.glob("programs/*.clvm"):
    hexname = fn[:-4] + "hex"
    if isfile(hexname):
        continue
    with open(hexname, "w+") as out:
        print(f"compiling {fn}")
        proc = subprocess.Popen(["opc", fn], stdout=out)
        proc.wait()

for fn in glob.glob("programs/*.env"):
    hexenv = fn + "hex"
    if isfile(hexenv):
        continue
    with open(hexenv, "w+") as out:
        print(f"compiling {fn}")
        proc = subprocess.Popen(["opc", fn], stdout=out)
        proc.wait()

for hexname in sorted(glob.glob("programs/*.hex")):
    hexenv = hexname[:-3] + "envhex"

    command = ["./run.py", hexname, hexenv]

    # prepend the size command, to measure RSS
    if platform.system() == "Darwin":
        command = ["/usr/bin/time", "-l"] + command
    if platform.system() == "Linux":
        command = ["/usr/bin/time"] + command

    print(" ".join(command))
    start = time.perf_counter()
    proc = subprocess.run(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    output = proc.stderr.decode("UTF-8")
    output += proc.stdout.decode("UTF-8")
    end = time.perf_counter()

    test = splitext(basename(hexname))[0]
    expected_error = expect[test]
    if f"FAIL: {expected_error}" not in output:
        ret += 1
        print(
            Fore.RED
            + f'\nTEST FAILURE: expected "{expected_error}"\n'
            + Style.RESET_ALL
        )
        print(output)

    print(Fore.YELLOW + ("  Runtime: %0.2f s" % (end - start)) + Style.RESET_ALL)

    # parse RSS (MacOS and Linux only)
    size = None
    if platform.system() == "Darwin":
        for l in output.split("\n"):
            if "maximum resident set size" not in l:
                continue
            val, key = l.strip().split("  ", 1)
            if key == "maximum resident set size":
                size = int(val) / 1024 / 1024
    if platform.system() == "Linux":
        # example output:
        # 10.49user 0.32system 0:10.84elapsed 99%CPU (0avgtext+0avgdata 189920maxresident)k
        for l in output.split("\n"):
            if "maxresident)k" not in l:
                continue
            size = int(l.split("maxresident)k")[0].split(" ")[-1]) / 1024
    if size != None:
        print(Fore.YELLOW + ("  Resident Size: %d MiB" % size) + Style.RESET_ALL)

        if size > 2300:
            ret += 1
            print(
                Fore.RED
                + "\nTEST FAILURE: Max memory use exceeded (limit: 2300 MB)\n"
                + Style.RESET_ALL
            )

    # cost 10923314721 roughly corresponds to 11 seconds
    if end - start > 11:
        ret += 1
        print(
            Fore.RED
            + "\nTEST FAILURE: Time exceeded: %f (limit: 11)\n" % (end - start)
            + Style.RESET_ALL
        )

if ret:
    print(Fore.RED + f"\n   There were {ret} failures!\n" + Style.RESET_ALL)

sys.exit(ret)



import unittest

from clvm_rs.program import Program


class castable_to_bytes:
    def __init__(self, blob):
        self.blob = blob

    def __bytes__(self):
        return self.blob


class dummy_class:
    pass


def gen_tree(depth: int) -> Program:
    if depth == 0:
        return Program.to(1337)
    subtree = gen_tree(depth - 1)
    return Program.to((subtree, subtree))


fh = bytes.fromhex
H01 = fh("01")
H02 = fh("02")


class AsPythonTest(unittest.TestCase):
    def test_int(self):
        v = Program.to(42)
        self.assertEqual(v.atom, bytes([42]))
        self.assertEqual(v.as_int(), 42)
        v = Program.to(0)
        self.assertEqual(v.atom, bytes([]))
        self.assertEqual(v.as_int(), 0)

    def test_empty_list(self):
        v = Program.to([])
        self.assertEqual(v.atom, b"")

    def test_list_of_one(self):
        v = Program.to([1])
        self.assertEqual(type(v.pair[0]), Program)
        self.assertEqual(type(v.pair[1]), Program)
        self.assertEqual(type(v.pair[0]), Program)
        self.assertEqual(type(v.pair[1]), Program)
        self.assertEqual(v.pair[0].atom, b"\x01")
        self.assertEqual(v.pair[1].atom, b"")

    def test_g1element(self):
        # we don't import `G1Element` here, but just replicate
        # its most important property: a `__bytes__` method
        b = fh(
            "b3b8ac537f4fd6bde9b26221d49b54b17a506be147347dae5"
            "d081c0a6572b611d8484e338f3432971a9823976c6a232b"
        )
        v = Program.to(castable_to_bytes(b))
        self.assertEqual(v.atom, b)

    def test_listp(self):
        self.assertEqual(Program.to(42).listp(), False)
        self.assertEqual(Program.to(b"").listp(), False)
        self.assertEqual(Program.to(b"1337").listp(), False)

        self.assertEqual(Program.to((1337, 42)).listp(), True)
        self.assertEqual(Program.to([1337, 42]).listp(), True)

    def test_nullp(self):
        self.assertEqual(Program.to(b"").nullp(), True)
        self.assertEqual(Program.to(b"1337").nullp(), False)
        self.assertEqual(Program.to((b"", b"")).nullp(), False)

    def test_constants(self):
        self.assertEqual(Program.null().nullp(), True)
        self.assertEqual(Program.null(), Program.to([]))
        self.assertEqual(Program.null(), Program.to(0))
        self.assertEqual(Program.one(), Program.to(1))

    def test_list_len(self):
        v = Program.to(42)
        for i in range(100):
            self.assertEqual(v.list_len(), i)
            v = Program.to((42, v))
        self.assertEqual(v.list_len(), 100)

    def test_list_len_atom(self):
        v = Program.to(42)
        self.assertEqual(v.list_len(), 0)

    def test_as_int(self):
        self.assertEqual(Program.to(fh("80")).as_int(), -128)
        self.assertEqual(Program.to(fh("ff")).as_int(), -1)
        self.assertEqual(Program.to(fh("0080")).as_int(), 128)
        self.assertEqual(Program.to(fh("00ff")).as_int(), 255)
        self.assertEqual(Program.fromhex("ff8080").as_int(), None)
        with self.assertRaises(ValueError):
            int(Program.fromhex("ff8080"))

    def test_string(self):
        self.assertEqual(Program.to("foobar").atom, b"foobar")

    def test_deep_recursion(self):
        d = b"2"
        for i in range(1000):
            d = [d]
        v = Program.to(d)
        for i in range(1000):
            self.assertEqual(v.pair[1].atom, Program.null())
            v = v.pair[0]
            d = d[0]

        self.assertEqual(v.atom, b"2")
        self.assertEqual(d, b"2")

    def test_long_linked_list(self):
        d = b""
        for i in range(1000):
            d = (b"2", d)
        v = Program.to(d)
        for i in range(1000):
            self.assertEqual(v.pair[0].atom, d[0])
            v = v.pair[1]
            d = d[1]

        self.assertEqual(v.atom, b"")
        self.assertEqual(d, b"")

    def test_long_list(self):
        d = [1337] * 1000
        v = Program.to(d)
        for i in range(1000):
            self.assertEqual(v.pair[0].as_int(), d[i])
            v = v.pair[1]

        self.assertEqual(v.atom, b"")

    def test_invalid_tuple(self):
        with self.assertRaises(ValueError):
            Program.to((dummy_class, dummy_class))

        with self.assertRaises(ValueError):
            Program.to((dummy_class, dummy_class, dummy_class))

    def test_clvm_object_tuple(self):
        o1 = Program.to(b"foo")
        o2 = Program.to(b"bar")
        self.assertEqual(Program.to((o1, o2)), (o1, o2))

    def test_first(self):
        val = Program.to(1)
        self.assertEqual(val.first(), None)
        val = Program.to((42, val))
        self.assertEqual(val.first(), Program.to(42))

    def test_rest(self):
        val = Program.to(1)
        self.assertEqual(val.first(), None)
        val = Program.to((42, val))
        self.assertEqual(val.rest(), Program.to(1))

    def test_as_iter(self):
        val = list(Program.to((1, (2, (3, (4, b""))))).as_iter())
        self.assertEqual(val, [1, 2, 3, 4])

        val = list(Program.to(b"").as_iter())
        self.assertEqual(val, [])

        val = list(Program.to((1, b"")).as_iter())
        self.assertEqual(val, [1])

        # these fail because the lists are not null-terminated
        self.assertEqual(list(Program.to(1).as_iter()), [])
        self.assertEqual(
            list(Program.to((1, (2, (3, (4, 5))))).as_iter()), [1, 2, 3, 4]
        )

    def test_eq(self):
        val = Program.to(1)

        self.assertEqual(val, 1)
        self.assertNotEqual(val, 2)

        # mismatching types
        self.assertNotEqual(val, [1])
        self.assertNotEqual(val, [1, 2])
        self.assertNotEqual(val, (1, 2))
        self.assertNotEqual(val, (dummy_class, dummy_class))

    def test_eq_tree(self):
        val1 = gen_tree(2)
        val2 = gen_tree(2)
        val3 = gen_tree(3)

        self.assertTrue(val1 == val2)
        self.assertTrue(val2 == val1)
        self.assertFalse(val1 == val3)
        self.assertFalse(val3 == val1)

    def test_str(self):
        self.assertEqual(str(Program.to(1)), "01")
        self.assertEqual(str(Program.to(1337)), "820539")
        self.assertEqual(str(Program.to(-1)), "81ff")
        self.assertEqual(str(gen_tree(1)), "ff820539820539")
        self.assertEqual(str(gen_tree(2)), "ffff820539820539ff820539820539")

    def test_repr(self):
        self.assertEqual(repr(Program.to(1)), "Program(01)")
        self.assertEqual(repr(Program.to(1337)), "Program(820539)")
        self.assertEqual(repr(Program.to(-1)), "Program(81ff)")
        self.assertEqual(repr(gen_tree(1)), "Program(ff820539820539)")
        self.assertEqual(repr(gen_tree(2)), "Program(ffff820539820539ff820539820539)")



"""
Serialize clvm.

# decoding:
# read a byte
# if it's 0xfe, it's nil (which might be same as 0)
# if it's 0xff, it's a cons box. Read two items, build cons
# otherwise, number of leading set bits is length in bytes to read size
# 0-0x7f are literal one byte values
# leading bits is the count of bytes to read of size
# 0x80-0xbf is a size of one byte (`and` of first byte with 0x3f for size)
# 0xc0-0xdf is a size of two bytes (`and` of first byte with 0x1f)
# 0xe0-0xef is 3 bytes (`and` of first byte with 0xf)
# 0xf0-0xf7 is 4 bytes (`and` of first byte with 0x7)
# 0xf7-0xfb is 5 bytes (`and` of first byte with 0x3)
"""

from typing import BinaryIO, Callable, Iterator, List

from .clvm_storage import CLVMStorage


MAX_SINGLE_BYTE = 0x7F
CONS_BOX_MARKER = 0xFF


def sexp_to_byte_iterator(sexp: CLVMStorage) -> Iterator[bytes]:
    """
    Yields bytes that serialize the given clvm object. Non-recursive
    """
    todo_stack = [sexp]
    while todo_stack:
        sexp = todo_stack.pop()
        r = getattr(sexp, "_cached_serialization", None)
        if r is not None:
            yield r
            continue
        pair = sexp.pair
        if pair:
            yield bytes([CONS_BOX_MARKER])
            todo_stack.append(pair[1])
            todo_stack.append(pair[0])
        else:
            atom = sexp.atom
            assert atom is not None
            yield from atom_to_byte_iterator(atom)


def size_blob_for_blob(blob: bytes) -> bytes:
    size = len(blob)
    if size < 0x40:
        return bytes([0x80 | size])
    if size < 0x2000:
        return bytes([0xC0 | (size >> 8), (size >> 0) & 0xFF])
    if size < 0x100000:
        return bytes([0xE0 | (size >> 16), (size >> 8) & 0xFF, (size >> 0) & 0xFF])
    if size < 0x8000000:
        return bytes(
            [
                0xF0 | (size >> 24),
                (size >> 16) & 0xFF,
                (size >> 8) & 0xFF,
                (size >> 0) & 0xFF,
            ]
        )
    if size < 0x400000000:
        return bytes(
            [
                0xF8 | (size >> 32),
                (size >> 24) & 0xFF,
                (size >> 16) & 0xFF,
                (size >> 8) & 0xFF,
                (size >> 0) & 0xFF,
            ]
        )
    raise ValueError("blob too long %r" % blob)


def atom_to_byte_iterator(as_atom: bytes) -> Iterator[bytes]:
    """
    Yield the serialization for a given blob (as a clvm atom).
    """
    size = len(as_atom)
    if size == 0:
        yield b"\x80"
        return
    if size == 1:
        if as_atom[0] <= MAX_SINGLE_BYTE:
            yield as_atom
            return
    yield size_blob_for_blob(as_atom)
    yield as_atom


def sexp_to_stream(sexp: CLVMStorage, f: BinaryIO) -> None:
    """
    Serialize to a file.
    """
    for b in sexp_to_byte_iterator(sexp):
        f.write(b)


def sexp_to_bytes(sexp: CLVMStorage) -> bytes:
    b = bytearray()
    for _ in sexp_to_byte_iterator(sexp):
        b.extend(_)
    return bytes(b)


NEW_PAIR_F = Callable[[CLVMStorage, CLVMStorage], CLVMStorage]
NEW_ATOM_F = Callable[[bytes], CLVMStorage]
OP_STACK_F = Callable[
    [List["OP_STACK_F"], List[CLVMStorage], BinaryIO, NEW_PAIR_F, NEW_ATOM_F], None
]


def _op_read_sexp(
    op_stack: List[OP_STACK_F],
    val_stack: List[CLVMStorage],
    f: BinaryIO,
    new_pair_f: NEW_PAIR_F,
    new_atom_f: NEW_ATOM_F,
):
    blob = f.read(1)
    if len(blob) == 0:
        raise ValueError("bad encoding")
    b = blob[0]
    if b == CONS_BOX_MARKER:
        op_stack.append(_op_cons)
        op_stack.append(_op_read_sexp)
        op_stack.append(_op_read_sexp)
        return
    val_stack.append(_atom_from_stream(f, b, new_atom_f))


def _op_cons(
    op_stack: List[OP_STACK_F],
    val_stack: List[CLVMStorage],
    f: BinaryIO,
    new_pair_f: NEW_PAIR_F,
    new_atom_f: NEW_ATOM_F,
):
    right = val_stack.pop()
    left = val_stack.pop()
    val_stack.append(new_pair_f(left, right))


def sexp_from_stream(
    f: BinaryIO, new_pair_f: NEW_PAIR_F, new_atom_f: NEW_ATOM_F
) -> CLVMStorage:
    op_stack: List[OP_STACK_F] = [_op_read_sexp]
    val_stack: List[CLVMStorage] = []

    while op_stack:
        func = op_stack.pop()
        func(op_stack, val_stack, f, new_pair_f, new_atom_f)
    return val_stack.pop()


def _atom_from_stream(f: BinaryIO, b: int, new_atom_f: NEW_ATOM_F) -> CLVMStorage:
    if b == 0x80:
        return new_atom_f(b"")
    if b <= MAX_SINGLE_BYTE:
        return new_atom_f(bytes([b]))
    bit_count = 0
    bit_mask = 0x80
    while b & bit_mask:
        bit_count += 1
        b &= 0xFF ^ bit_mask
        bit_mask >>= 1
    size_blob = bytes([b])
    if bit_count > 1:
        blob = f.read(bit_count - 1)
        if len(blob) != bit_count - 1:
            raise ValueError("bad encoding")
        size_blob += blob
    size = int.from_bytes(size_blob, "big")
    if size >= 0x400000000:
        raise ValueError("blob too large")
    blob = f.read(size)
    if len(blob) != size:
        raise ValueError("bad encoding")
    return new_atom_f(blob)
